@@ -140,18 +140,6 @@ class CharArg(ArgType):
 	info.varlist.add('gchar', 'ret')
         info.codeafter.append('    return PyString_FromStringAndSize(&ret, 1);')
 class GUniCharArg(ArgType):
-    param_tmpl = ('    if (py_%(name)s[1] != 0) {\n'
-                  '        PyErr_SetString(PyExc_TypeError, "%(name)s should be a 1 character unicode string");\n'
-                  '        return NULL;\n'
-                  '    }\n'
-                  '    %(name)s = (gunichar)py_%(name)s[0];\n')
-    dflt_tmpl = ('    if (py_%(name)s != NULL) {\n'
-                 '        if (py_%(name)s[1] != 0) {\n'
-                 '            PyErr_SetString(PyExc_TypeError, "%(name)s should be a 1 character unicode string");\n'
-                 '            return NULL;\n'
-                 '        }\n'
-                 '        %(name)s = (gunichar)py_%(name)s[0];\n'
-                 '     }\n')
     ret_tmpl = ('#if !defined(Py_UNICODE_SIZE) || Py_UNICODE_SIZE == 2\n'
                 '    if (ret > 0xffff) {\n'
                 '        PyErr_SetString(PyExc_RuntimeError, "returned character can not be represented in 16-bit unicode");\n'
@@ -163,13 +151,10 @@ class GUniCharArg(ArgType):
     def write_param(self, ptype, pname, pdflt, pnull, info):
 	if pdflt:
 	    info.varlist.add('gunichar', pname + " = '" + pdflt + "'")
-            info.codebefore.append(self.dflt_tmpl % {'name':pname})
 	else:
 	    info.varlist.add('gunichar', pname)
-            info.codebefore.append(self.param_tmpl % {'name':pname})
-        info.varlist.add('Py_UNICODE', '*py_' + pname + ' = NULL')
 	info.arglist.append(pname)
-        info.add_parselist('u', ['&py_' + pname], [pname])
+        info.add_parselist('O&', ['pyg_pyobj_to_unichar_conv', '&' + pname], [pname])
     def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('gunichar', 'ret')
         info.varlist.add('Py_UNICODE', 'py_ret')
@@ -333,7 +318,7 @@ class EnumArg(ArgType):
         info.add_parselist('O', ['&py_' + pname], [pname]);
     def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('gint', 'ret')
-        info.codeafter.append('    return PyInt_FromLong(ret);')
+        info.codeafter.append('    return pyg_enum_from_gtype(%s, ret);' % self.typecode)
 
 class FlagsArg(ArgType):
     flag = ('    if (%(default)spyg_flags_get_value(%(typecode)s, py_%(name)s, (gint *)&%(name)s))\n'
@@ -356,7 +341,7 @@ class FlagsArg(ArgType):
         info.add_parselist('O', ['&py_' + pname], [pname])
     def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('guint', 'ret')
-        info.codeafter.append('    return PyInt_FromLong(ret);')
+        info.codeafter.append('    return pyg_flags_from_gtype(%s, ret);' % self.typecode)
 
 class ObjectArg(ArgType):
     # should change these checks to more typesafe versions that check
@@ -417,7 +402,8 @@ class ObjectArg(ArgType):
         if ownsreturn:
             info.varlist.add('PyObject', '*py_ret')
             info.codeafter.append('    py_ret = pygobject_new((GObject *)ret);\n'
-                                  '    g_object_unref(ret);\n'
+                                  '    if (ret != NULL)\n'
+                                  '        g_object_unref(ret);\n'
                                   '    return py_ret;')
         else:
             info.codeafter.append('    /* pygobject_new handles NULL checking */\n' +
